/* debugger.c : tout pour mettre au point */

#include "mlvalues.h"
#include "stacks.h"
#include "debugger.h"
#include "instruct.h"
#include "opnames.h"
#include "memory.h"
#include "unaligned.h"

code log_buffer[LOG_BUFFER_SIZE];
code * log_ptr;

int trace_flag;

/* Affichage d'un objet structure */

long max_print = 100;
long max_print_depth = 10;

long print_cnt;

static void print_val(v, d)
	value v;
	long d;
{
  long n;
  value * p;

  if (d <= 0) {
    printf(".");
    return;
  }
  print_cnt--;
  if (print_cnt <= 0) {
    if (print_cnt == 0) printf("...");
    return;
  }
  if (IS_INT(v))
    printf("%d", CINT(v));
  else if (!IS_ALLOCATED(v))
    printf("0x%x", v);
  else switch(TAG(v)) {
    case TAG_STRING:
      printf("\"%s\"", CSTRING(v));
      break;
    case TAG_FLOAT:
      printf("%g", CFLOAT(v));
      break;
    case TAG_BITVECT:
      printf("<bitvect>");
      break;
    default:
      n = TAG(v);
      if (n < 26)
        printf("%c", n + 'A');
      else
        printf("tag%d", n);
      n = SIZE(v);
      if (n > 0) {
        printf("(");
        p = &FIELD(v, 0);
        while (n > 1) {
          print_val(*p, d-1);
          printf(", ");
          p++;
          n--;
        }
        print_val(*p, d-1);
        printf(")");
      }
      break;
  }
}

void print_value(v)
	value v;
{
  print_cnt = max_print;
  print_val(v, max_print_depth);
  printf("\n");
}

extern code start_code;

void print_pc(pc)
     code pc;
{
  printf("0x%6d  ", pc - start_code);
}

/* Desassemblage d'une instruction */

code disasm_instr(pc)
	code pc;
{
  int i;

  print_pc(pc);
  i = *pc++;
  if (i < 0 || i >= sizeof(names_of_instructions) / sizeof(char *)) {
    printf("???\n");
    return pc;
  }
  printf("%s ", names_of_instructions[i]);
  switch(i) {
/* instructions avec parametre numerique sur 1 octet */
    case ACCESS: case DUMMY: case UPDATE: case ENDLET: case CONSTBYTE:
    case ATOM: case GETFIELD: case SETFIELD:
    case MAKEBLOCK1: case MAKEBLOCK2: case MAKEBLOCK3: case MAKEBLOCK4: 
    case C_CALL1: case C_CALL2: case C_CALL3: case C_CALL4: case C_CALL5:
    case C_CALL_ALLOC1: case C_CALL_ALLOC2: case C_CALL_ALLOC3:
    case C_CALL_ALLOC4: case C_CALL_ALLOC5:
      printf("%d", *pc++);
      break;
/* instruction avec parametre numerique sur 2 octets, non signe */
    case GETGLOBAL: case SETGLOBAL:
      printf("%d", u16(pc)); pc += 2; break;
/* instruction avec parametre numerique sur 2 octets, signe */
    case CONSTSHORT: 
      printf("%d", s16(pc)); pc += 2; break;
/* instruction avec parametre numerique sur 4 octets */
    case CONSTLONG: case MAKEBLOCK:
      printf("%d", s32(pc)); pc += 4; break;
/* instructions avec deplacement */
    case CUR: case LETREC1: case PUSHTRAP:
    case BRANCH: case BRANCHIF: case BRANCHIFNOT:
    case BRANCHIFEQ: case BRANCHIFNEQ: case BRANCHIFLT:
    case BRANCHIFGT: case BRANCHIFLE: case BRANCHIFGE:
    depl:
      { int orig = pc - start_code;
	printf("%d", orig + s16(pc));
	pc += 2;
        break;
      }
/* instructions avec tag + deplacement */
    case BRANCHIFEQTAG: case BRANCHIFNEQTAG:
      printf("tag %d, ", (unsigned) *pc++);
      goto depl;
/* instructions diverses */
    case FLOATOP:
      printf("%s", names_of_float_instructions[*pc++]);
      break;
    case SWITCH:
#define SWITCH_DEFAULT_SIZE 8
      { code orig = pc;
	int i;
	for (i = 0; i < SWITCH_DEFAULT_SIZE - 1; i++) {
	  printf("%d, ", orig + s16(pc) - start_code);
          pc += 2;
	}
        printf("%d, ", orig + s16(pc) - start_code);
        return orig;
      }
    case BRANCHIFNEQIMM:
      printf("%d", s32(pc)); pc += 4; break;
    case BRANCHIFNEQIMMSTRING:
      { int orig = pc - start_code;
	int c;
	orig = orig + s16(pc);
        pc += 2;
	while(c = *pc++) putchar(c);
	printf(", %d", orig);
	break;
      }
    case STARTFUN:
      pc += 2;
      break;
    }
  printf("\n");
  return pc;
}

void disasm(pc)
	code pc;
{
  int i;

  for (i = 0; i < 20; i++)
    pc = disasm_instr(pc);
}

void post_mortem(n)
	int n;
{
  code * p;

  if (n > LOG_BUFFER_SIZE) n = LOG_BUFFER_SIZE;
  for (p = log_buffer +
             (unsigned) (log_ptr - log_buffer - n) % LOG_BUFFER_SIZE;
       n > 0;
       n--) {
    disasm_instr(*p);
    p++;
    if (p >= log_buffer + LOG_BUFFER_SIZE) p = log_buffer;
  }
}
