
/* interp.c  :  l'interpreteur de bytecode, version C rapide */

#include <math.h>
#include <setjmp.h>
#include "mlvalues.h"
#include "mem_misc.h"
#include "stacks.h"
#include "memory.h"
#include "cprim.h"
#include "instruct.h"
#include "alloc.h"
#include "fail.h"
#include "unaligned.h"
#include "signals.h"
#include "strings.h"
#ifdef DEBUG
#include "debugger.h"
#endif
#ifdef macintosh
#include <CursorCtl.h>
static value spin_count = 0;
#endif

/* Les registres de la machine abstraite */

/*	pc  	le pointeur de programme
	asp  	le pointeur de pile d'arguments (croit vers le bas)
	rsp  	le pointeur de pile de retour (croit vers le bas)
	tp  	le pointeur vers le trap_frame actif
	env  	la partie remanente (dans le tas) de l'environnement
     cache_size le nombre de valeurs volatiles dans l'environnement
	accu  	l'accumulateur

"asp" et "rsp" sont des contreparties locales de "extern_asp" et "extern_rsp".

*/

extern value global_data;
extern jmp_buf * external_raise;
extern code start_code;

/******************** Les macros de l'interprete ***********************/

/* D'autres point de vue sur rsp */

#define retsp  ((struct return_frame *) rsp)
#define trapsp ((struct trap_frame   *) rsp)

#define push_ret_frame()  (rsp -= sizeof(struct return_frame) / sizeof(value))
#define pop_ret_frame()   (rsp += sizeof(struct return_frame) / sizeof(value))
#define push_trap_frame() (rsp -= sizeof(struct trap_frame) / sizeof(value))
#define pop_trap_frame()  (rsp += sizeof(struct trap_frame) / sizeof(value))

/* D'autres points de vue sur le pc (pour lire les operandes immediats). */

#define SHORT  (sizeof(short))
#define LONG   (sizeof(long))
#define DOUBLE (sizeof(double))

#define c8pc  (* (char *) pc)
#ifdef INTERP_ALIGN
#include "unaligned.h"
#define s16pc s16(pc)
#define u16pc u16(pc)
#define s32pc s32(pc)
#define u32pc u32(pc)
#else
#define s16pc (* (short *) (pc))
#define u16pc (* (unsigned short *) (pc))
#define s32pc (* (long *) (pc))
#define u32pc (* (unsigned long *) (pc))
#endif

/* L'environnement vide */

#define null_env atom(0)

/* Le code "raise Bottom" */

instruction code_for_dummy[] = { ATOM0 + BOTTOM_EXN, RAISE };

/* pour sauver les registres avant d'appeler le gc. */

#undef SETUP_FOR_GC
#undef RESTORE_AFTER_GC

#define SETUP_FOR_GC							      \
  { push_ret_frame();							      \
    retsp->env = env;							      \
    retsp->cache_size = cache_size;					      \
    *--asp = accu;							      \
    extern_asp = asp; extern_rsp = rsp;					      \
  }

#define RESTORE_AFTER_GC						      \
  { accu = *asp++;							      \
    env = retsp->env;							      \
    pop_ret_frame ();							      \
  }

/* Pareil, mais sans sauver l'accu.
   A faire avant l'appel d'une primitive C qui peut allouer. */

#define SETUP_FOR_C_CALL                                                      \
  { push_ret_frame();                                                         \
    retsp->env = env;                                                         \
    retsp->cache_size = cache_size;                                           \
    extern_asp = asp;                                                         \
    extern_rsp = rsp;                                                         \
  }
#define RESTORE_AFTER_C_CALL                                                  \
  { asp = extern_asp;                                                         \
    rsp = extern_rsp;                                                         \
    env = retsp->env;                                                         \
    pop_ret_frame();                                                          \
  }

/* Pour mettre l'environnement tout entier dans le tas */

#define heapify_env()							      \
{									      \
  int env_size = SIZE(env);						      \
  int new_size = env_size + cache_size;					      \
  value * from, * to;							      \
									      \
  SMALL_ALLOC(tmp,  MAKE_HEADER(new_size,0), new_size);			      \
  for(to = POINTER(tmp); cache_size > 0; cache_size--) *to++ = *rsp++;	      \
  for(from = POINTER(env); env_size > 0; env_size--  ) *to++ = *from++;	      \
  env = tmp;								      \
}

/* L'interprete en personne */

value interprete(prog, start)          /* ML */
     code prog;
     value start;
{
/* Les declarations des registres de la machine abstraite.
   Donnes par frequence d'utilisation decroissante.
   Suivant les architectures et les compilateurs, on peut avoir a enlever
   certains "register", ou a imposer des registres (GCC).
   En particulier, la plupart des compilateurs optimiseurs sous-estiment
   l'importance de "pc", qui DOIT etre mis en registre. */

  register code pc              /* Le pointeur de code */
#ifdef __GNUC__
#ifdef mc68000
                   asm("a5")
#endif
#ifdef sparc
                   asm("%l1")
#endif
#endif
                            ;
  register value accu;          /* L'accumulateur */
  register value * asp;		/* Contrepartie locale de extern_asp*/
  register value * rsp;		/* ... de extern_rsp */
  int cache_size;               /* Nombre d'entrees dans l'env. volatile */
  value env;                    /* L'environnement persistant */
  value tmp;		        /* Auxiliaire */
  jmp_buf * initial_external_raise;
  				/* "external_raise" au debut de l'interprete */
  int initial_rsp_offset;       /* "rsp" au demarrage de l'interprete */
  jmp_buf raise_buf;            /* Pour traiter les d'exceptions externes */

  asp = extern_asp;
  rsp = extern_rsp;
  pc = prog + CINT(start);
  env = null_env;
  cache_size = 0;
  accu = MLINT(0);

  if (setjmp(raise_buf)) {
    accu = exn_bucket;
    goto raise;
  }
  initial_external_raise = external_raise;
  external_raise = (jmp_buf *) raise_buf;
  initial_rsp_offset = ret_stack_high - rsp;

#ifdef DEBUG
  log_ptr = log_buffer;
#endif

  while (1) {

#ifdef DEBUG
    *log_ptr++ = pc;
    if (log_ptr >= log_buffer + LOG_BUFFER_SIZE)
      log_ptr = log_buffer;
    if (trace_flag)
      disasm_instr(pc);
    Zassert(asp >= arg_stack_low);
    Zassert(asp <= arg_stack_high);
    Zassert(rsp >= ret_stack_low);
    Zassert(rsp <= ret_stack_high);
#endif DEBUG

    switch (*pc++) {

    case STOP:
      extern_asp = asp;
      extern_rsp = rsp;
      external_raise = initial_external_raise;
      return accu;
      
/********** Construction et application de valeurs fonctionnelles **********/
      
    case CUR:
      if (cache_size) heapify_env();
      SMALL_ALLOC(accu, HEADER_CLOSURE, SIZE_CLOSURE);
      ENV(accu) = env;
      CODE(accu) = pc + s16pc;
      pc += SHORT;
      break;
      
    case APPLY:
      push_ret_frame();
      retsp->pc = pc;
      retsp->env = env;
      retsp->cache_size = cache_size;
      *--rsp = *asp++;		/* Place le premier arg. dans l'env. */
      cache_size = 1;
      pc = CODE(accu);
      env = ENV(accu);
      break;
      
    case STARTFUN:
      if (asp < arg_stack_threshold || rsp < ret_stack_threshold) {
        SETUP_FOR_GC;
        realloc_stacks();
        rsp = extern_rsp;
        asp = extern_asp;
        RESTORE_AFTER_GC;
      }
      pc += SHORT;
      if (signal_is_pending) goto raise_signal;
#ifdef macintosh
	  /* This is the right time to be Multifinder-friendly. */
	  if ((spin_count++ & 15) == 0) SpinCursor ((short) 1);
#endif
      break;

    case RETURN:
      if (*asp == MARK) {
	rsp += cache_size;	/* Vide le cache */
	asp++;			/* Detruit la marque */
	pc = retsp->pc;		/* Et retourne a l'appelant */
	env = retsp->env;
	cache_size = retsp->cache_size;
	pop_ret_frame();
	break;
      }
      /* else fall through APPTERM */
      
    case APPTERM:
      rsp += cache_size;	/* Vide le cache */
      *--rsp = *asp++;		/* Place le premier arg. dans l'env. */
      cache_size = 1;
      pc = CODE(accu);
      env = ENV(accu);
      break;
      
/******* Manipulations de l'environnement (acces, ajout, destruction) *****/
      
    case GRAB:
      if (*asp != MARK) {
	*--rsp = *asp++;
	cache_size++;
      } else {
	if (cache_size) heapify_env();
	SMALL_ALLOC(accu, HEADER_CLOSURE, SIZE_CLOSURE);
	CODE(accu) = pc;
	ENV(accu) = env;
	asp++;			/* Detruit la marque */
	pc = retsp->pc;
	env = retsp->env;
	cache_size = retsp->cache_size;
	pop_ret_frame();
      }
      break;
      
#define access(n) (cache_size > n ? rsp[n] : FIELD(env, n - cache_size))
#define access0() (cache_size > 0 ? rsp[0] : FIELD(env,0))
      /* Un peu plus efficace que access(0)... */

    case ACC0:
      accu = access0(); break;
    case ACC1:
      accu = access(1); break;
    case ACC2:
      accu = access(2); break;
    case ACC3:
      accu = access(3); break;
    case ACC4:
      accu = access(4); break;
    case ACC5:
      accu = access(5); break;
    case ACCESS:
      { int n = *pc++;
	accu = access(n);
	break;
      }
      
    case LET:
      *--rsp = accu;
      cache_size++;
      break;
      
    case DUMMY:
      { int n = *pc++;
	Zassert (n > 0);
	SMALL_ALLOC(tmp, MAKE_HEADER (n, 0), n);
	while (n--){
	  FIELD (tmp, n) = MLINT (0);
	}
	*--rsp = tmp;
	cache_size++;
	break;
      }
      
    case UPDATE:
      { int n = *pc++;
	tmp = access(n);
	TAG (tmp) = TAG (accu);
	for (n = 0; n < SIZE (accu); n++){
	  MODIFY (&FIELD (tmp, n), FIELD (accu, n));
	}
	break;
      }
      
    case LETREC1:		/* Remplace Dummy 1; Cur lbl; Update 0 */
      SMALL_ALLOC(accu, HEADER_CLOSURE, SIZE_CLOSURE);
      FIELD(accu,0) = FIELD(accu,1) = atom(0);
				/* Necessaire si GC pendant heapify_env */
      *--rsp = accu;
      cache_size++;
      heapify_env();
      accu = FIELD(env,0);	/* Necessaire si GC pendant heapify_env */
      CODE(accu) = pc + s16pc;  /* Si code non collectable */
      MODIFY(&ENV(accu), env);	/* On boucle l'environnement */
      pc += SHORT;
      break;
      
    case ENDLET1:
      if (cache_size != 0) {
	cache_size--; rsp++;
      } else {
	int i;                  /* On recopie dans le cache l'ancien env. */
	value * from;           /* prive de sa premiere valeur */
	i = SIZE(env);
        from = &FIELD(env, i);
        cache_size = i - 1;
        for (i = cache_size; i > 0; i--) *--rsp = *--from;
	env = null_env;
      }
      break;
      
    case ENDLET:
      { int n = *pc++;
	if (cache_size >= n) {
	  cache_size -= n;
	  rsp += n;
	} else {
	  int i;
	  value * from;
	  n -= cache_size;	/* Ce qu'il faut oter dans l'env du tas */
	  rsp += cache_size;	/* On vide entierement le cache */
	  			/* On recopie ds le cache l'ancien env */
	  			/* prive de ses n premieres valeurs */
          i = SIZE(env);
	  cache_size = i - n;
          from = &FIELD(env, i);
	  for (i = cache_size; i > 0; i--) *--rsp = *--from;
	  env = null_env;
	}
	break;
      }
      
/**************** Gestion des exceptions ******************/
      
    case PUSHTRAP:
      { value * src = rsp + cache_size;
	int i = cache_size;
	
	push_trap_frame();		/* On sauve l'etat de la machine */
	trapsp->pc = pc + s16pc;
	pc += SHORT;
	trapsp->env = env;
	trapsp->cache_size = cache_size + 2;  /* +2 : astuce pour le GC */
	trapsp->asp = asp;
	trapsp->tp = tp;
	tp = trapsp;
	while(i--) *--rsp = *--src;	/* On duplique le cache */
	*--asp = MARK;		/* On stoppe la consommation d'arguments */
	break;
      }

    raise_signal:               /* La detection d'un signal amene ici */
      accu = pending_signal;
      signal_is_pending = 0;
      pending_signal = MLINT(1);

    raise:			/* un raise externe amene ici */

    case RAISE:
      rsp = (value *) tp;	/* On arase la pile R */
      if (rsp >= ret_stack_high - initial_rsp_offset) {
                                /* On sort de l'interprete? */
        exn_bucket = accu;
        external_raise = initial_external_raise;
        longjmp(*external_raise, 1);
      }
      pc = trapsp->pc;		/* On restaure l'etat de la machine */
      env = trapsp->env;
      cache_size = trapsp->cache_size - 2;   /* -2 : astuce pour le GC */
      asp = trapsp->asp;
      tp = trapsp->tp;
      pop_trap_frame();		/* On depile le trap frame */
      *--rsp = accu;		/* On met le bucket en tete de l'env. */
      cache_size++;
      break;
      
    case POPTRAP:
      if (signal_is_pending) goto raise_signal;
      rsp = (value *) tp;
      env = trapsp->env;
      cache_size = trapsp->cache_size - 2;
      asp = trapsp->asp;
      tp = trapsp->tp;
      pop_trap_frame();
      break;
      
/***************** Chargement de constantes dans l'accu *******************/
      
    case CONSTBYTE:		/* Pour les petits entiers */
      accu = c8pc;  pc += 1; break;
    case CONSTSHORT:		/* Pour les moins petits entiers */
      accu = s16pc; pc += SHORT; break;
    case CONSTLONG:		/* Pour une valeur hors du tas quelconque */
      accu = s32pc; pc += LONG; break;
    case ATOM0:
      accu = atom(0); break;
    case ATOM1:
      accu = atom(1); break;
    case ATOM2:
      accu = atom(2); break;
    case ATOM3:
      accu = atom(3); break;
    case ATOM4:
      accu = atom(4); break;
    case ATOM5:
      accu = atom(5); break;
    case ATOM6:
      accu = atom(6); break;
    case ATOM7:
      accu = atom(7); break;
    case ATOM8:
      accu = atom(8); break;
    case ATOM9:
      accu = atom(9); break;
    case ATOM:
      accu = atom(*pc++); break;
      
/********** Acces aux valeurs globales *****************/
      
    case GETGLOBAL:
      accu = FIELD(global_data, u16pc);
      pc += SHORT;
      break;
    case SETGLOBAL:
      MODIFY(&FIELD(global_data, u16pc), accu);
      pc += SHORT;
      break;
      
/************ Les classique manipulations de pile (d'arguments) **********/
      
    case PUSH:
      *--asp = accu; break;
    case POP:
      accu = *asp++; break;
    case PUSHMARK:
      *--asp = MARK;
      break;
      
/************ Branchements, branchements conditionnels et tutti quanti *******/
      
    case NOP1:			/* Generes pour assurer l'alignement */
      break;			/* si necessaire */
    case NOP2:
      pc++; break;
    case NOP3:
      pc += 2; break;
      
#define branch() pc += s16pc
#define cond_branch(condition) if (condition) branch(); else pc += 2
      
    case BRANCH:
      branch(); break;
    case BRANCHIF:
      cond_branch(TAG(accu) != 0); break;
    case BRANCHIFNOT:
      cond_branch(TAG(accu) == 0); break;
    case BRANCHIFEQTAG:
      if (TAG(accu) == *pc++) branch(); else pc += SHORT;
      break;
    case BRANCHIFNEQTAG:
      if (TAG(accu) != *pc++) branch(); else pc += SHORT;
      break;
    case SWITCH:
      pc += s16(pc + accu - 1);
      break;
    case BOOLNOT:
      accu = atom(TAG(accu) == 0); break;
      
/******************* Acces aux champs d'un bloc **********************/
      
    case GETFIELD0:
      accu = FIELD(accu,0); break;
    case GETFIELD1:
      accu = FIELD(accu,1); break;
    case GETFIELD2:
      accu = FIELD(accu,2); break;
    case GETFIELD3:
      accu = FIELD(accu,3); break;
    case GETFIELD:
      accu = FIELD(accu,*pc++); break;
      
    case SETFIELD0:
      tmp = accu; accu = *asp++; MODIFY(&FIELD(tmp,0), accu); break;
    case SETFIELD1:
      tmp = accu; accu = *asp++; MODIFY(&FIELD(tmp,1), accu); break;
    case SETFIELD2:
      tmp = accu; accu = *asp++; MODIFY(&FIELD(tmp,2), accu); break;
    case SETFIELD3:
      tmp = accu; accu = *asp++; MODIFY(&FIELD(tmp,3), accu); break;
    case SETFIELD:
      { value * ptr;
        ptr = &FIELD(accu,*pc++);
        accu = *asp++;
        MODIFY(ptr, accu);
        break;
      }
      
/******************* Construction d'un bloc **************************/
      
    case MAKEBLOCK1:
      SMALL_ALLOC(tmp, MAKE_HEADER(1, *pc), 1);
      pc++;
      FIELD(tmp,0) = accu;
      accu = tmp;
      break;
    case MAKEBLOCK2:
      SMALL_ALLOC(tmp, MAKE_HEADER(2, *pc), 2);
      pc++;
      FIELD(tmp,0) = accu;
      FIELD(tmp,1) = *asp++;
      accu = tmp;
      break;
    case MAKEBLOCK3:
      SMALL_ALLOC(tmp, MAKE_HEADER(3, *pc), 3);
      pc++;
      FIELD(tmp,0) = accu;
      FIELD(tmp,1) = *asp++;
      FIELD(tmp,2) = *asp++;
      accu = tmp;
      break;
    case MAKEBLOCK4:
      SMALL_ALLOC(tmp, MAKE_HEADER(4, *pc), 4);
      pc++;
      FIELD(tmp,0) = accu;
      FIELD(tmp,1) = *asp++;
      FIELD(tmp,2) = *asp++;
      FIELD(tmp,3) = *asp++;
      accu = tmp;
      break;
    case MAKEBLOCK:
      { header hdr;
	blocksize size;
	value * to;
	
	hdr = u32pc;
	pc += LONG;
	size = SIZE_HDR(hdr);
        if (size < MAX_YOUNG_SIZE) {
          SMALL_ALLOC(tmp, hdr, size);
          to = &FIELD(tmp, 0);
          *to++ = accu;
          for (size--; size > 0; size--) *to++ = *asp++;
          accu = tmp;
        } else {
          SETUP_FOR_GC;
          accu = big_alloc_init_up(hdr, size);
          asp = extern_asp;
          RESTORE_AFTER_GC;
        }
	break;
      }
      
    case TAGOF:
      accu = MLINT(TAG(accu));
      break;

/******************* Appel de fonctions C ***********************/
      
    case C_CALL1:
      accu = (cprim[*pc++])(accu);
      break;
    case C_CALL2:
      accu = (cprim[*pc++])(accu, asp[0]);
      asp += 1;
      break;
    case C_CALL3:
      accu = (cprim[*pc++])(accu, asp[0], asp[1]);
      asp += 2;
      break;
    case C_CALL4:
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2]);
      asp += 3;
      break;
    case C_CALL5:
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2], asp[3]);
      asp += 4;
      break;
    case C_CALLN:
      { int n = *pc++;
        *--asp = accu;
        accu = (cprim[*pc++])(asp, n);
        asp += n;
        break; }

    case C_CALL_ALLOC1:
      SETUP_FOR_C_CALL;
      accu = (cprim[*pc++])(accu);
      RESTORE_AFTER_C_CALL;
      break;
    case C_CALL_ALLOC2:
      SETUP_FOR_C_CALL;
      accu = (cprim[*pc++])(accu, asp[0]);
      RESTORE_AFTER_C_CALL;
      asp += 1;
      break;
    case C_CALL_ALLOC3:
      SETUP_FOR_C_CALL;
      accu = (cprim[*pc++])(accu, asp[0], asp[1]);
      RESTORE_AFTER_C_CALL;
      asp += 2;
      break;
    case C_CALL_ALLOC4:
      SETUP_FOR_C_CALL;
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2]);
      RESTORE_AFTER_C_CALL;
      asp += 3;
      break;
    case C_CALL_ALLOC5:
      SETUP_FOR_C_CALL;
      accu = (cprim[*pc++])(accu, asp[0], asp[1], asp[2], asp[3]);
      RESTORE_AFTER_C_CALL;
      asp += 4;
      break;
    case C_CALL_ALLOCN:
      { int n = *pc++;
        *--asp = accu;
        SETUP_FOR_C_CALL;
        accu = (cprim[*pc++])(asp, n);
        RESTORE_AFTER_C_CALL;
        asp += n;
        break; }
      
/******************** Primitives sur les entiers **********************/
      
    case NEGINT:
      accu = 2 - accu; break;
    case SUCCINT:
      accu += 2; break;
    case PREDINT:
      accu -= 2; break;
    case ADDINT:
      accu += *asp++ - 1; break;
    case SUBINT:
      accu -= *asp++ - 1; break;
    case MULINT:
      accu = 1 + (accu >> 1) * (*asp++ - 1); break;
    case DIVINT:
      tmp = *asp++ - 1;
      if (tmp == 0) {
        accu = atom(ZERO_DIVIDE_EXN);
        goto raise;
      }
      accu = MLINT((accu - 1) / tmp);
      break;
    case MODINT:
      tmp = *asp++ - 1;
      if (tmp == 0) {
        accu = atom(ZERO_DIVIDE_EXN);
        goto raise;
      }
      accu = 1 + (accu - 1) % tmp;
      break;
    case ANDINT:
      accu &= *asp++; break;
    case ORINT:
      accu |= *asp++; break;
    case XORINT:
      accu = 1 + (accu ^ *asp++); break;
    case SHIFTLEFTINT:
      accu = 1 + ((accu - 1) << CINT(*asp++)); break;
    case SHIFTRIGHTINT:
      accu = 1 | ((accu - 1) >> CINT(*asp++)); break;
      
#define inttest(name1,name2,tst)					      \
    case name1:								      \
      accu = atom(accu tst *asp++);					      \
      break;								      \
    case name2:								      \
      cond_branch(accu tst *asp++);					      \
      break;
      
      inttest(EQ, BRANCHIFEQ, ==);
      inttest(NEQ, BRANCHIFNEQ, !=);
      inttest(LTINT, BRANCHIFLT, <);
      inttest(GTINT, BRANCHIFGT, >);
      inttest(LEINT, BRANCHIFLE, <=);
      inttest(GEINT, BRANCHIFGE, >=);
      
    case BRANCHIFNEQIMM:
      if (accu != s32pc) { pc += LONG; branch(); } else pc += LONG+SHORT;
      break;
    case INCR:
      accu = (FIELD(accu, 0) += 2); break;
    case DECR:
      accu = (FIELD(accu, 0) -= 2); break;

/****************** Primitives sur les flottants *************************/
      
    case FLOATOP:
      { SMALL_ALLOC(tmp, HEADER_FLOAT, SIZE_FLOAT);
	switch(*pc++) {
	case FLOATOFINT: CFLOAT(tmp) = CINT(accu); break;
	case NEGFLOAT:   CFLOAT(tmp) = -CFLOAT(accu); break;
	case ADDFLOAT:   CFLOAT(tmp) = CFLOAT(accu) + CFLOAT(*asp++); break;
	case SUBFLOAT:   CFLOAT(tmp) = CFLOAT(accu) - CFLOAT(*asp++); break;
	case MULFLOAT:   CFLOAT(tmp) = CFLOAT(accu) * CFLOAT(*asp++); break;
	case DIVFLOAT:   CFLOAT(tmp) = CFLOAT(accu) / CFLOAT(*asp++); break;
	}
	accu = tmp;
	break;
      }
      
    case INTOFFLOAT:
      accu = MLINT((long)CFLOAT(accu)); break;
      
#define floattest(name, tst)    					      \
    case name:								      \
      accu = atom(CFLOAT(accu) tst CFLOAT(*asp++));			      \
      break;
      
      floattest(EQFLOAT, ==);
      floattest(NEQFLOAT, !=);
      floattest(LTFLOAT, <);
      floattest(GTFLOAT, >);
      floattest(LEFLOAT, <=);
      floattest(GEFLOAT, >=);
      
    case BRANCHIFNEQIMMFLOAT:
      if (CFLOAT(accu) != *((double *)pc))
        { pc += DOUBLE; branch(); }
      else
        pc += DOUBLE+SHORT;
      break;

/********************* Primitives sur les chaines ***********************/
      
    case MAKESTRING:
      { blocksize size = CINT(accu);
        ALLOC_STRING(accu, size);
        break; }
    case STRINGLENGTH:
      { blocksize size;
        GET_STRING_LENGTH(accu, size);
        accu = MLINT(size);
        break; }
    case GETSTRINGCHAR:
      accu = MLINT(BYTE(accu, CINT(*asp++)));
      break;
    case SETSTRINGCHAR:
      BYTE(accu, CINT(asp[0])) = CINT(asp[1]);
      accu = asp[1];
      asp += 2;
      break;

    case BRANCHIFNEQIMMSTRING:
      { code newpc = pc + s16pc;
        unsigned char * p = CSTRING(accu);
        pc += SHORT;
        do {
          if (*p++ != *pc) { pc = newpc; break; }
        } while (*pc++);
        break;
      }
      
#define stringtest(name, tst)                                                 \
    case name:                                                                \
      accu = atom(compare_strings(accu, *asp++) tst MLINT(0));                \
      break;
      
      stringtest(EQSTRING, ==);
      stringtest(NEQSTRING, !=);
      stringtest(LTSTRING, <);
      stringtest(GTSTRING, >);
      stringtest(LESTRING, <=);
      stringtest(GESTRING, >=);
      
/********************* Primitives sur les vecteurs *********************/
      
    case MAKEVECTOR:
      { unsigned long size = CINT(accu);
        if (size == 0)
          accu = atom(0);
        else
          ALLOC_INIT(accu, MAKE_HEADER(size, 0), size, *asp);
        asp++;
	break;
      }
    case VECTLENGTH:
      accu = MLINT(SIZE(accu));
      break;
    case GETVECTITEM:
      accu = FIELD(accu, CINT(*asp++));
      break;
    case SETVECTITEM:
      tmp = asp[1];
      MODIFY(& FIELD(accu, CINT(asp[0])), tmp);
      asp += 2;
      accu = tmp;
      break;
    }
  }
}
